function [mu,S2SR,S2PP,rmse] =gprSRPPv(logtheta,covfunc,x,INDEX,y,xstar,ystar);

% gprSRPPv - Carries out approximate Gaussian process regression prediction
% using the subset of regressors (SR) or projected process approximation (PP)
% with the active set specified by INDEX.  This version uses a variation
% of the Peters-Wilkinson approach for solving least squares problems to 
% increase the numerical stability of the algorithm.
%
% Usage alternatives:
%
%   mu = gprSRPPv(logtheta, covfunc, X, INDEX, y, Xstar);
%   [mu, rmse] = gprSRPPv(logtheta, covfunc, X, INDEX, y, Xstar);
%   [mu, S2SR, S2PP] = gprSRPPv(logtheta, covfunc, X, INDEX, y, Xstar);
%   [mu, S2SR, S2PP, rmse] = gprSRPPv(logtheta, covfunc, X, INDEX, y, Xstar);
%   [mu, rmse] = gprSRPPv(logtheta, covfunc, X, INDEX, y, Xstar, ystar);
%   [mu, S2SR, S2PP, rmse] = gprSRPPv(logtheta,covfunc,X,INDEX,y,Xstar, ystar);
%
% where
%
% Input (ystar is optional):
%   logtheta a (column) vector of log hyperparameters
%   covfunc  the covariance function, which is assumed to
%            be a covSum, and the last entry of the sum is covNoise
%   x        a n by D matrix of training inputs
%   INDEX    a vector of length m <= n used to specify which 
%            inputs are used in the active set.  Note that in some 
%            cases (if a Cholesky factorization fails) m is reduced to 
%            insure that the factorization succeeds. In this case the 
%            active set is INDEX(1:m).
%   y        a (column) vector (of size n) of targets
%   xstar    a nstar by D matrix of test inputs
%   ystar    is a (column) vector (of size nstar) of test targets
% Output (S2SR, S2PP and rmse are optional):
%   mu       a (column) vector (of size nstar) of predicted means
%   S2SR     a (column) vector (of size nstar) of predicted variances under SR
%   S2PP     a (column) vector (of size nstar) of predicted variances under PP
%   rmse     a (column) vector (of size m) of root mean square errors
%            If the number of input arguments is 6:
%                rmse(i) = || y_est - y || / sqrt(n), i = 1,...,m
%            where y_est is the estimated value of y with a rank i approx.
%            If the number of input arguments is 7:
%                rmse(i) = || y_pred - ystar || / sqrt(nstar), i = 1,...,m
%            where y_pred is the predicted value of ystar with a rank i approx.
%            Note that length(rmse) gives the current (potentially reduced) m.
%
% For more help on covariance functions, see "help covFunctions".

% (C) modification (2008-3-23) by Leslie Foster of code by Chris Williams.

% Notes:  memory use can be reduced by changing the npartition = 0 line below.
%         memory and time can be reduced by changing rank_increment =1 below.

if ischar(covfunc), covfunc = cellstr(covfunc); end % convert to cell if needed
[n, D] = size(x);
if eval(feval(covfunc{:})) ~= size(logtheta, 1)
  error('Error: Number of parameters do not agree with covariance function')
end

% we check that the covfunc cell array is a covSum, with last entry 'covNoise'
if length(covfunc) ~= 2 | ~strcmp(covfunc(1), 'covSum') | ...
                                           ~strcmp(covfunc{2}(end), 'covNoise')
  error('The covfunc must be "covSum" whose last summand must be "covNoise"')
end

sigma2n = exp(2*logtheta(end));                                % noise variance
[nstar, D] = size(xstar);   % number of test cases and dimension of input space
m = length(INDEX);                                             % size of subset

% note, that in the following Kmm is computed by extracting the relevant part
% of Knm, thus it will be the "noise-free" covariance (although the covfunc
% specification does include noise).
[v, Knm] = feval(covfunc{:}, logtheta, x, x(INDEX,:)); 
Kmm = Knm(INDEX,:);                     % Kmm is a noise-free covariance matrix

[V11,p]=chol(Kmm); 
if ( p ~= 0 )
   %  necessary if Kmm is (numerically) not positive definite
   %  first the code tries perturbing Kmm and, if this fails, m is reduced
   if ( 1 == 0 )   % replace by "if ( 1 == 0 )" to skip this altenative
      jitter = 1.0e-9 * trace(Kmm);       % as suggested in code of r\w and jqc
      Kmm = Kmm + jitter * eye(m,m);
      [V11,p] = chol(Kmm);
   end
   if ( p ~= 0 )
      %  reduce the rank since the Kmm block does not have a nonsing. chol 
      %  factor.  Also select the 1st m entries of INDEX to be the active set.
      m = p-1;
      Knm = Knm(:,1:m);
      INDEX = INDEX(1:m);
   end
end

Knm_t_y = Knm'*y;                                  % used to calculate mu below

npartition = 0;          % npartition = 0 to maximize speed but use more memory  
%npartition = 10000;   % is a reasonable choice if memory is a concern
if ( npartition == 0 )
   %Knm = Knm/V11;              % Matlab requires extra storage to execute this 
   Knm = Knm * inv(V11);         % a small bit slower but less storage required
else
   nstart = 1;
   nstop = npartition;
   while (nstart <= n) 
      Knm(nstart:nstop,:) = Knm(nstart:nstop,:)/V11;
      nstart = nstop +1;
      nstop = min(nstop+npartition, n );
   end
end

V11=V11';                                         % V11 is now lower triangular

[L,p] = chol(sigma2n*eye(m,m)+ Knm' * Knm);
if ( p ~= 0 )
   %  necessary if (sigma2n*eye(m,m) + Knm'*Knm) is not SPD (numerically)
   %  reduce rank since (sigma2n*eye(m,m) + Knm'*Knm) does not have a nonsing. 
   %  chol factor. Select the 1st m entries of INDEX for the active set.
   m = p-1;
   Knm = Knm(:,1:m);
   V11 = V11(1:m,1:m);
   INDEX = INDEX(1:m);
   Knm_t_y = Knm_t_y(1:m);
end
L = L';                                             % L is now lower triangular

% K1star is cov between active set and test points and vstar is variances at
% test points, incl noise variance

[vstar, K1star] = feval(covfunc{:}, logtheta, x(INDEX,:), xstar); 
K1star = K1star';        % K1star is now cov between test points and active set 

mu = K1star * (V11' \ ( L' \ ( L \ (V11 \(Knm_t_y) )) ) ); 
      
if ( nargout >= 3 )
   % calculate variances 
   K1star = K1star / V11' ;                   % overwrite K1star to save memory
   S2PP = vstar - sum(K1star .* K1star, 2 );
   K1star = ( K1star / L' );                  % overwrite K1star to save memory
   S2SR = sigma2n * sum(K1star .* K1star,2); % noise-free SR variance, eq. 8.15
   S2PP = S2PP + S2SR;                 % PP variance eq. (8.27) including noise
   S2SR = S2SR + sigma2n ;                        % SR variance inclusing noise
end
         
if ( nargout == 2 | nargout == 4 )
   % calculate rmse errors for lower rank approximations
   vt = L \ ( V11 \ Knm_t_y );
   
   B = triu( vt*ones(1,m) );
   rank_increment = 1;         % rank_increment = 1 for rmse at all lower ranks
   %rank_increment = 10;     % rank_increment =10 (say) to save time and memory
   if ( rank_increment == 1 )
       rankv = 1:m;
   else
       % the rmse errors will be calculated only at the ranks in rankv
       rankv = 1:rank_increment:m;
       if (rankv(end) ~= m ), rankv = [rankv, m]; end
       B = B(:,rankv);
   end
   
   if ( nargin == 6 )
      % calculate rmse errors for estimated y, y_est
      Knm = Knm * (  L' \  B );                    % cols of Knm now have y_est
      sqrt_n = sqrt(n);
      for i = 1:length(rankv)
         rmse(i) = norm(y - Knm(:,i)) / sqrt_n; 
      end
   end
   
   if ( nargin == 7 )
      % calculate rmse errors for predicted ystar, y_pred
      if ( nargout <= 2 )
         if ( nstar >= m )
            K1star = K1star * ( V11' \ ( L' \  B ) );
         else
            K1star = ( ( K1star / V11' ) / L' ) * B;
         end
      else
         % make use of earlier computations from variance calculation
         K1star = K1star *  B;
      end
      % columns of K1star now have y_pred
      sqrt_nstar = sqrt(nstar);
      for i = 1:length(rankv)
         rmse(i) = norm(ystar - K1star(:,i)) / sqrt_nstar;
      end
   end
   
   if ( nargout == 2 )
      S2SR = rmse;
   end
   
end
